@@ -1,23 +1,21 @@
-from descope.auth import Auth
+from typing import Optional
+
+from descope._auth_base import AuthBase
 from descope.common import EndpointsV1, LoginOptions, validate_refresh_token_provided
 from descope.exceptions import ERROR_TYPE_INVALID_ARGUMENT, AuthException
 
 
-class SAML:
-    _auth: Auth
-
-    def __init__(self, auth: Auth):
-        self._auth = auth
-
+# This class is DEPRECATED please use SSO instead
+class SAML(AuthBase):
     def start(
         self,
         tenant: str,
-        return_url: str = None,
-        login_options: LoginOptions = None,
-        refresh_token: str = None,
+        return_url: Optional[str] = None,
+        login_options: Optional[LoginOptions] = None,
+        refresh_token: Optional[str] = None,
     ) -> dict:
         """
-        Docs
+        DEPRECATED
         """
         if not tenant:
             raise AuthException(
